from .MeetingTimeEstimator import MeetingTimeEstimator

import numpy as np
import networkx as nx
from scipy import sparse
from typing import Tuple


class MeetingTimesUI:
    """

    This class provides a user interface for the code computing the structural diversity index.

    """

    @staticmethod
    def _get_random_walk_simulator_networkx(g: nx.Graph, on_cuda: bool):
        """

        Given a graph it gets the random walk simulator for that graph.

        Args:
            g (nx.Graph): the graph on which you want to simulate a random walk
            on_cuda (bool): if True, uses cuda to make computations

        Returns:
            (RandomWalkSimulator or RandomWalkSimulatorCUDA): An instance of the class RandomWalkSimulator or RandomWalkSimulatorCUDA

        """
        adjacency = nx.to_scipy_sparse_array(G=g, format='csr', dtype=float)
        degrees = np.array([d[1] for d in g.degree])
        return MeetingTimesUI._get_random_walk_simulator(adjacency=adjacency, degrees=degrees, on_cuda=on_cuda)

    @staticmethod
    def _get_random_walk_simulator(adjacency: sparse.csr_matrix, degrees: np.ndarray, on_cuda: bool):
        """

        Given an adjacency matrix and a degree sequence of a graph it gets the random walk simulator for that graph.

        Args:
            adjacency (sparse.csr_matrix): the adjacency matrix of the graph
            degrees (np.ndarray): the degree sequence of the graph
            on_cuda (bool): if True, uses cuda to make computations

        Returns:
            (RandomWalkSimulator or RandomWalkSimulatorCUDA): An instance of the class RandomWalkSimulator or RandomWalkSimulatorCUDA

        """
        if on_cuda:
            from .RandomWalkSimulatorCUDA import RandomWalkSimulator as RandomWalkSimulatorCUDA
            random_walk_matrix = RandomWalkSimulatorCUDA.random_walk_matrix_from_adjacency(adjacency=adjacency,
                                                                                           degrees=degrees)
            rws = RandomWalkSimulatorCUDA(random_walk_matrix=random_walk_matrix)
        else:
            from .RandomWalkSimulator import RandomWalkSimulator
            random_walk_matrix = RandomWalkSimulator.random_walk_matrix_from_adjacency(adjacency=adjacency,
                                                                                       degrees=degrees)
            rws = RandomWalkSimulator(random_walk_matrix=random_walk_matrix)
        return rws

    @staticmethod
    def get_meeting_times(g: nx.Graph, max_time_steps: int = None, n_samples: int = 1000, on_cuda: bool = False) -> np.ndarray:
        """

        Returns the n_samples of the meeting time of two uniformly started random walks on the graph g


         Args:
             g (nx.Graph): the graph for which you want the meeting time samples
             max_time_steps (int): the maximum number of time steps to run the random walks \
              to compute the meeting time (default = 10*g.number_of_nodes())
             n_samples (int): the number of samples of the meeting time (default = 1000)
             on_cuda (bool): if True, uses cuda to make computations (default = False)

         Returns:
             (np.ndarray): An array in which each entry is one sample of the meeting time on the graph g. \
             Note: To estimate the meeting time of random walks that do not meet within max_times_steps steps \
             we use the method MeetingTimeEstimator.estimate_meeting_times_unmet_random_walks(). \
             See the class for a description.

         """

        if max_time_steps is None:
            max_time_steps = 10 * g.number_of_nodes()

        # With k random walks we can compute k**2 - k samples of the meeting time. Hence, if we want
        # n_samples of the meeting time we just need [1 + sqrt(1+4n_samples)]/2 random walks
        n_samples_random_walks = int(np.ceil((1 + np.sqrt(1+4*n_samples))/2))

        # Create a random walk simulator and compute the meeting times
        rws = MeetingTimesUI._get_random_walk_simulator_networkx(g=g, on_cuda=on_cuda)
        meeting_times = rws.get_meeting_times(max_time_steps=max_time_steps, n_samples=n_samples_random_walks)
        meeting_times = MeetingTimeEstimator.estimate_meeting_times_unmet_random_walks(meeting_times=meeting_times,
                                                                                       max_time_steps=max_time_steps)

        # We get some more samples of the meeting time because n_samples_random_walks is a bit larger than [1 + sqrt(1+4n_samples)]/2
        # Hence we cut out some samples
        meeting_times = meeting_times[:n_samples]

        return meeting_times

    @staticmethod
    def get_structural_diversity_index(g: nx.Graph, max_time_steps: int = None, n_samples: int = 100, on_cuda: bool = False) -> Tuple[float, float]:
        """

        Returns the structural diversity index of a graph g

         Args:
             g (nx.Graph): the graph for which you want the structural diversity index
             max_time_steps (int): the maximum number of time steps to run the random walks \
              to compute the meeting time (default = 10*g.number_of_nodes())
             n_samples (int): the number of samples of the meeting time (default = 1000)
             on_cuda (bool): if True, uses cuda to make computations (default = False)

         Returns:
             (float, float): The structural diversity index of the graph g and the standard deviation in the samples \
             used for the computation. This index is computed by obtaining n_samples of the meeting time of the graph g \
              and averaging over them. Note: To estimate the meeting time of random walks that do not meet \
              within max_times_steps steps we use the method MeetingTimeEstimator.estimate_meeting_times_unmet_random_walks(). \
             See the class for a description.

         """

        meeting_times = MeetingTimesUI.get_meeting_times(g=g, max_time_steps=max_time_steps, n_samples=n_samples, on_cuda=on_cuda)
        delta_g = float(np.mean(meeting_times) / g.number_of_nodes())
        std_delta_g = float(np.std(meeting_times/g.number_of_nodes()))
        return delta_g, std_delta_g



import numpy as np


class MeetingTimeEstimator:
    """

    The class MeetingTimeEstimator implements some methods that allow to estimate the value of a meeting time sample, given that the walks simulated to compute that sample did not meet.
    The main idea behind these methods is to use the fact that meeting times are approximately geometrically distributed

    """

    @staticmethod
    def estimate_meeting_times_unmet_random_walks(meeting_times: np.ndarray, max_time_steps: int) -> np.ndarray:
        """

        Given samples of the meeting times of two random walks containing some samples in which the walks did not meet, this method estimates the value of the samples in which the walks did not meet.

        Our estimate for the meeting time is based on the following reasoning. If the meeting time is approximately geometrically distributed it suffices to:
            * Estimate the parameter p of the geometric distribution which best approximate the distribution of the meeting time
            * Replace each -1 (i.e. each sample of the meeting time to be estimated) with: v + max_time_steps + 1, where v is a value sampled from the geometric distribution with parameter p. \
              Here we are using the memory-less property of the geometric distribution.


        Args:
            meeting_times (1D np.ndarray): a 1D np.ndarray containing samples of the meeting times of two random walks. Indicate by -1 the samples which you want to be estimated
            max_time_steps (int): The number of time steps that the random walks have been run in order to compute the samples of the meeting times passed in the array meeting_times. \
            For the samples with a -1 we assume that the walks took more than max_time_steps to meet.

        Returns:
             (1D np.ndarray): a 1D np.ndarray containing samples of the meeting times of two random walks. The -1 in the array passed as parameter are replaced by estimates for the meeting time.

        """

        # If there are no walks that did not meet there is nothing to estimate
        n_unmet_random_walks = len(meeting_times) - np.count_nonzero(meeting_times + 1)
        if n_unmet_random_walks == 0:
            return meeting_times

        # Otherwise, we estimate the meeting time as described above
        else:
            # Estimate the parameter p of the geometric distribution
            p = MeetingTimeEstimator.estimate_p_geometric_approximation(meeting_times=meeting_times,
                                                                        max_time_steps=max_time_steps)

            # Find the indices of the samples where the walks did not meet
            index_unmet_walks = np.argwhere(meeting_times == -1).reshape(-1)

            # Replace the value -1 at the indices in index_unmet_walks with the estimate constructed as described above
            meeting_times_with_estimate = np.zeros(len(meeting_times))
            meeting_times_with_estimate[index_unmet_walks] = np.random.geometric(p=p, size=len(
                index_unmet_walks)) + max_time_steps + 1
            meeting_times_with_estimate = meeting_times_with_estimate + meeting_times

            return meeting_times_with_estimate

    @staticmethod
    def estimate_p_geometric_approximation(meeting_times: np.ndarray, max_time_steps: int) -> float:
        """

        Computes the parameter p of the geometric distribution which best approximate the distribution of the meeting time.

        The computation is done by using the following equation:
        P[T leq t] = (1 - p)^t.
        Here T is the meeting time of the walks. If T is geometrically distributed the equation above should hold for some p.
        To compute this p, we compute all terms in the equation and then isolate p.
        Here t = max_time_steps and P[T leq t] is computed by counting the number of walks that did not meet.

        Args:
            meeting_times (1D np.ndarray): a 1D np.ndarray containing samples of the meeting times of two random walks. Indicate by -1 the samples which you want to be estimated
            max_time_steps (int): The number of time steps that the random walks have been run in order to compute the samples of the meeting times passed in the array meeting_times. \
            For the samples with a -1 we assume that the walks took more than max_time_steps to meet.

        Returns:
             (float): the parameter p of the geometric distribution which best approximate the distribution of the meeting time
        """

        # Compute P[T \leq t] where T is the meeting time of the random walks and t = max_time_steps
        n_unmet_random_walks = len(meeting_times) - np.count_nonzero(meeting_times + 1)
        probability_not_met = n_unmet_random_walks / len(meeting_times)

        # Compute p using the formula P[T \leq t] = (1 - p)^t
        p = 1 - np.exp(np.log(probability_not_met) / max_time_steps)

        return p



import numpy as np
import scipy.sparse as sparse
from tqdm import tqdm
import torch
from typing import Tuple


class RandomWalkSimulator:
    """

    The class RandomWalkSimulator is designed to run a fast simulations of a random walk on a graph
    and compute the meeting times of two walks

    """

    def __init__(self, random_walk_matrix: sparse.csr_matrix):
        """

        Initialises a RandomWalkSimulator

        Args:
            random_walk_matrix (scipy.sparse.csr_matrix): the random walk transition matrix of the graph on which you want to simulate the random walk

        """

        # Device name
        self.n_nodes = random_walk_matrix.shape[0]

        # Random walk matrix
        self.P = random_walk_matrix

    ###################################################################### PUBLIC METHODS ###############################################################################

    def get_meeting_times(self, max_time_steps: int, n_samples: int) -> np.ndarray:
        """

        Gets the meeting times necessary to compute delta_g, i.e., it compute n_samples of the meeting time of two randomly started walks.

        Args:
            max_time_steps (int): the number of time steps for which you want to simulate the random walks at most
            n_samples (int): the number of samples of the meeting time that you want

        Returns:
            (1D np.ndarray): a 1D np.ndarray in which each entry is one sample of the meeting time of two randomly started walks. If the walks met after max_time_steps the entry is equal to -1 by default.

        """
        start_position = np.random.randint(low=0, high=self.n_nodes, size=[n_samples])
        meeting_times = self._get_meeting_times(max_time_steps=max_time_steps, start_position=start_position)

        meeting_times_flat = np.ndarray.flatten(meeting_times)
        meeting_times_flat_without_diagonal = np.delete(meeting_times_flat,
                                                        range(0, len(meeting_times_flat), len(meeting_times_flat) + 1),
                                                        0)

        return meeting_times_flat_without_diagonal

    @staticmethod
    def random_walk_matrix_from_adjacency(adjacency: sparse.csr_matrix, degrees: np.ndarray) -> sparse.csr_matrix:
        """

        Returns the transition matrix of the random walk

        Args:
            adjacency (scipy.sparse.csr_matrix): the adjacency matrix of the graph for which we want to build the transition matrix
            degrees (np.ndarray): the degrees of the nodes in the graph

        Returns:
            (scipy.sparse.csr_matrix):  the transition matrix of the random walk. This is defined as P(i,j) = 1/deg(i)

        """

        # Build the transition matrix
        A = sparse.csr_matrix(adjacency)

        one_over_degrees = np.array(1 / degrees)
        ind = np.arange(one_over_degrees.shape[0])
        D = sparse.csr_matrix((one_over_degrees, (ind, ind)), shape=(ind.shape[0], ind.shape[0]))

        P = D * A
        return P

    ###################################################################### PRIVATE METHODS ###############################################################################

    def _get_meeting_times(self, max_time_steps: int, start_position: np.ndarray) -> np.ndarray:
        """

        This method simulates random walks started at start_position and keeps track of their meeting times.
        Then it returns these meeting times, once all the walks have met or max_time_steps has expired.

        Args:
            max_time_steps (int): the number of time steps for which you want to simulate the random walks at most
            start_position (1D np.ndarray, optional): a 1D np.ndarray in which each entry is the starting positions of one sample of the random walk.

        Returns:
            (2D np.ndarray): a 2D np.ndarray with entry m,n being the meeting time between sample walk m and sample walk n. \
            If two walks never meet the value of the m,n entry is -1

        """

        n_samples = len(start_position)

        meeting_times = -1 * np.ones([n_samples, n_samples])

        # Fix the starting position and check if walks meet at the starting position
        start_pos, array_start = self._start_position(n_samples=n_samples, start_position=start_position)
        meeting_bool = self._check_meetings(current_pos=start_pos, meeting_times=meeting_times)
        meeting_times = meeting_times + meeting_bool

        # Run the walks and at each time step check if some walks have met. Also, end the loop if all walks have met.
        # Here a while condition would look better but we could not use tqdm to time it
        array_current = array_start
        for t in tqdm(range(1, max_time_steps)):
            # Find next position of the walks
            next_pos, array_next = self._next_step(array_current=array_current)
            array_current = array_next

            # Check meetings at this round
            meeting_bool = self._check_meetings(current_pos=next_pos, meeting_times=meeting_times)
            meeting_times = meeting_times + (t + 1) * meeting_bool

            # Verify if all walks have met
            complete = self._check_complete(meeting_times=meeting_times)
            if complete:
                return meeting_times

        return meeting_times

    @staticmethod
    def _check_meetings(current_pos: np.ndarray, meeting_times: np.ndarray) -> np.ndarray:
        """


        This method checks if the walks meet for the first time at the current step.

        Args:
            current_pos (1D np.ndarray): the current position of the walks (i.e. the names of the vertices at which the walks are)
            meeting_times (2D np.ndarray): the matrix of meeting times. Entry i, j is the meeting time of walk i with walk j.\
            The default value of -1 is set when the walks have not met yet.

        Returns:
            (2D np.ndarray of boolean values): A matrix with entry (i,j) being true if walk i has met walk j for the first time at current_position \
            and false otherwise (i.e. either the walks are not at the same position, or they have met before).

        """

        # Check if the walks are at the same position
        auxiliary_1 = current_pos.reshape(-1, 1) * np.ones((current_pos.shape[0], current_pos.shape[0]))
        same_pos_bool = (current_pos == auxiliary_1)

        # Check if the walks have met before
        auxiliary_2 = -1 * np.ones((meeting_times.shape[0], meeting_times.shape[1]))
        not_already_met_bool = (meeting_times == auxiliary_2)

        # If walk i,j are at the same place and have never met before set entry i,j = True, else set entry i,j = False.
        first_meeting_bool = same_pos_bool * not_already_met_bool

        return first_meeting_bool

    @staticmethod
    def _check_complete(meeting_times: np.ndarray) -> bool:
        """

        This method check if all the walks have met

        Args:
            meeting_times (2D np.ndarray): the matrix of meeting times. Entry i, j is the meeting time of walk i with walk j. \
            The default value of -1 is set when the walks have not met yet.

        Returns:
            (bool):  True if all the walks have met and False otherwise

        """

        # Compares an auxiliary matrix of all -1 with the meeting time matrix
        auxiliary = -1 * np.ones((meeting_times.shape[0], meeting_times.shape[1]))
        not_already_met_bool = (meeting_times == auxiliary)

        # Counts how many non zero entries are in the matrix not_already_met_bool
        nb_not_met = int(np.count_nonzero(not_already_met_bool))

        if nb_not_met == 0:
            return True
        else:
            return False

    def _next_step(self, array_current: sparse.csr_matrix) -> Tuple[np.ndarray, sparse.csr_matrix]:
        """

         Given the current step of all the samples of the random walk, computes the next step for all the samples of the random walk

        Args:
            array_current (sparse.csr_matrix): A matrix with rows indicating the current position of the random walk. \
            Each row is of the form [0,0, ..., 0, 1, 0, ..., 0]. The entry 1 indicates the current position of the random walk.

        Returns:
            (np.ndarray, scipy.sparse.csr_matrix): \
            A tuple consisting of:
                * A 1D np.ndarray with for each sample of the random walk the index of the vertex where the rw will jump next
                * A 2D scipy.sparse.csr_matrix with rows indicates the next position of the random walk of a sample.

            Each row is of the form [0,0, ..., 0, 1, 0, ..., 0]. The entry 1 indicates the next position of the random walk.

        """

        # Compute the probabilities of moving to given neighbors by multiplying sparse matrices current_step and P
        proba_next_pos = array_current * self.P

        # Sample index of next step of the random walk.
        next_pos = self._compute_next_position_from_proba(proba_next_pos)

        # Create indices for sparse matrix
        ind_ptr = np.arange(array_current.shape[0] + 1, dtype=np.int64)

        # Create values for sparse matrix
        values = np.ones(array_current.shape[0], dtype=np.float64)

        # Construct a 2D sparse tensor with rows of the form [0,0, ..., 0, 1, 0, ..., 0], where the entry 1 indicates the position of the rw
        array_next = sparse.csr_matrix((values, next_pos, ind_ptr),
                                       shape=(array_current.shape[0], array_current.shape[1]))

        # One could initialise the sparse matrix as follows (more intuitive)
        # array_next = sparse.csr_matrix((values, (ind_row, next_pos)), shape=(array_current.shape[0],array_current.shape[1]))
        # However, this slows down the code remarkably

        return next_pos, array_next

    @staticmethod
    def _compute_next_position_from_proba(proba_next_pos: sparse.csr_matrix) -> np.ndarray:
        """

        For each row i of proba_next_pos, corresponding to one random walk, it samples the column index j with the probability specified in the entry proba_next_pos[i][j].
        The index j corresponds to the next position of the walk

        Args:
            proba_next_pos (2D scipy.sparse.csr_matrix): a matrix with entry i,j being the probability of moving from i to j for the random walk. \
            The matrix has axis 0 of length n_samples (i.e. number of walks) and axis 1 of length n_nodes (i.e. the probability \
            for the walk to move at each one of those nodes)

        Returns:
            (1D np.ndarray): the next position of the walks, randomly sampled from proba_next_pos

        """

        # This method is actually 3xfaster if only we could write the cuda code for it!!
        # Indeed, instead of transforming proba_next_pos into an array with a bunch of zeros, we could only
        # consider the entries in which we are interested.

        # Changes a numpy array into a torch tensor
        proba_next_pos_torch = torch.as_tensor(proba_next_pos.toarray(), device='cpu')

        # Samples stuff using torch
        next_pos_torch = torch.multinomial(proba_next_pos_torch, 1).flatten()

        # Converts back to numpy format
        next_pos = next_pos_torch.numpy()

        return next_pos

    def _start_position(self, n_samples: int, start_position: np.ndarray) -> Tuple[np.ndarray, sparse.csr_matrix]:
        """

        Randomly chooses starting vertices for each sample of the random walks

        Args:
            n_samples (int): the number of random walks you want to sample
            start_position (1D np.ndarray): starting positions for each sample of the random walk

        Returns:
            [(np.ndarray, scipy.sparse.csr_matrix)]:
            A tuple consisting of:
                * A 1D np.ndarray with for each sample of the random walks the index of the starting vertex of the random walks
                * A 2D scipy.sparse.csr_matrix with rows indicates the starting position of the random walk of a sample.

            Each row is of the form [0,0, ..., 0, 1, 0, ..., 0]. The entry 1 indicates the position of the random walk.

        """

        start_pos = start_position

        # Create indices for sparse matrix
        ind_ptr = np.arange(n_samples + 1, dtype=np.int64)

        # Create values for sparse matrix
        values = np.ones(n_samples, dtype=np.float64)

        # Construct a 2D sparse tensor with rows of the form [0,0, ..., 0, 1, 0, ..., 0], where the entry 1 indicates the position of the rw
        array_start = sparse.csr_matrix((values, start_pos, ind_ptr), shape=(n_samples, self.n_nodes))

        # One could initialise the sparse matrix as follows (more intuitive)
        # array_next = sparse.csr_matrix((values, (ind_row, next_pos)), shape=(array_current.shape[0],array_current.shape[1]))
        # However, this slows down the code remarkably

        return start_pos, array_start




